;;; navi2ch-article.el --- article view module for navi2ch

;; Copyright (C) 2000 by 2$B$A$c$s$M$k(B

;; Author: (not 1)
;; Keywords: network, 2ch

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(require 'cl)
(require 'navi2ch-message)
(require 'browse-url)

(defvar navi2ch-article-enable-diff t
  "$B:9J,$r<h$C$F$/$k$+$I$&$+!#(Bnil $B$J$i>o$K:9J,$r<h$C$F$3$J$$(B")

(defvar navi2ch-article-enable-fill nil
  "fill-region $B$9$k$+$I$&$+(B")
(defvar navi2ch-article-enable-fill-list nil
  "fill-region $B$9$k(B $BHD$N%j%9%H(B")
(defvar navi2ch-article-disable-fill-list nil
  "fill-region $B$7$J$$HD$N%j%9%H(B")

(defvar navi2ch-article-enable-through t
  "$B%9%l%C%I$N:G8e$G%9%Z!<%9$r2!$7$?$H$-$K<!$N%9%l%C%I$K0\F0$9$k$+!#(B")

(defvar navi2ch-article-link-face 'bold)
(defvar navi2ch-article-link-mouse-face 'highlight)
(defvar navi2ch-article-url-face navi2ch-article-link-face)
(defvar navi2ch-article-url-mouse-face navi2ch-article-link-mouse-face)

(defvar navi2ch-article-parse-field-list
  '(data name mail)
  "parse $B$9$k%U%#!<%k%I$N%j%9%H!#(B
$BCY$/$F$b$$$$$s$J$i(B '(data mail name) $B$H$+$9$k$H$$$$$+$b(B")

(defvar navi2ch-article-goto-number-recenter t
  "goto-number $B$7$?$H$-$K(B recenter $B$9$k$+$I$&$+(B")

(defvar navi2ch-article-view-range nil
  "$BI=<($9$k%9%l%C%I$NHO0O!#(B
$B=q<0$O(B '(first . last) $B$G!"(B
first $B$,:G=i$+$i$$$/$DI=<($9$k$+!"(B
last $B$,:G8e$+$i$$$/$DI=<($9$k$+!#(B
$BNc$($P!"(B(10 . 50) $B$G!":G=i$N(B10$B$H:G8e$N(B50$B$rI=<((B")

(defvar navi2ch-article-new-message-range
  '(50 . 1)
  "$B?7$7$$(B $B%9%l%C%I$r<h$C$F$-$?$H$-$NI=<($9$kHO0O(B")
(defvar navi2ch-article-exist-message-range
  '(1 . 50)
  "$B$9$G$K$"$k%9%l%C%I$r<h$C$F$-$?$H$-$NI=<($9$kHO0O(B")
(defvar navi2ch-article-auto-range t
  "$B$^$?I=<($7$F$J$$%9%l%C%I$r<h$C$F$-$?$H$-$K>!<j$KHO0O$r69$a$k$+!#(B
non-nil $B$G69$a$k(B")

(defvar navi2ch-article-view-range-list
  '((1 . 50)
    (50 . 50)
    (1 . 100)
    (100 . 100))
  "$BI=<($9$k%9%l%C%I$NHO0O$rA*Br$9$k$H$-$K;H$&%j%9%H(B")
  
(defvar navi2ch-article-header-format-function
  'navi2ch-article-default-header-format-function
  "NUMBER NAME MAIL DATE $B$r0z?t$K<h$j!"%l%9$N%X%C%@$rJV$94X?t(B")

(defvar navi2ch-article-mode-map nil)
(unless navi2ch-article-mode-map
  (setq navi2ch-article-mode-map (make-sparse-keymap))
  (define-key navi2ch-article-mode-map "q" 'navi2ch-article-exit)
  (define-key navi2ch-article-mode-map " " 'navi2ch-article-scroll-up)
  (define-key navi2ch-article-mode-map "s" 'navi2ch-article-sync)
  (define-key navi2ch-article-mode-map "S" 'navi2ch-article-sync-disable-diff)
  (define-key navi2ch-article-mode-map "r" 'navi2ch-article-redraw-range)
  (define-key navi2ch-article-mode-map [del] 'navi2ch-article-scroll-down)
  (define-key navi2ch-article-mode-map [backspace] 'navi2ch-article-scroll-down)
  (define-key navi2ch-article-mode-map "\177" 'navi2ch-article-scroll-down)
  (define-key navi2ch-article-mode-map "w" 'navi2ch-article-write-message)
  (define-key navi2ch-article-mode-map "t" 'navi2ch-toggle-offline)
  (define-key navi2ch-article-mode-map "\r" 'navi2ch-article-select-current-link)
  (define-key navi2ch-article-mode-map [mouse-2] 'navi2ch-article-mouse-select)
  (define-key navi2ch-article-mode-map "g" 'navi2ch-article-goto-number)
  (define-key navi2ch-article-mode-map "l" 'navi2ch-article-pop-point)
  (define-key navi2ch-article-mode-map "m" 'navi2ch-article-push-point)
  (define-key navi2ch-article-mode-map "u" 'navi2ch-article-show-url)
  
  (define-key navi2ch-article-mode-map "." 'navi2ch-article-redisplay-current-message)
  (define-key navi2ch-article-mode-map "p" 'navi2ch-article-previous-message)
  (define-key navi2ch-article-mode-map "n" 'navi2ch-article-next-message)
  (define-key navi2ch-article-mode-map "P" 'navi2ch-article-through-prev)
  (define-key navi2ch-article-mode-map "N" 'navi2ch-article-through-next)
  (define-key navi2ch-article-mode-map [(shift tab)] 'navi2ch-article-previous-link)
  (define-key navi2ch-article-mode-map "\e\C-i" 'navi2ch-article-previous-link)
  (define-key navi2ch-article-mode-map "\C-\i" 'navi2ch-article-next-link)
  (define-key navi2ch-article-mode-map ">" 'navi2ch-article-goto-last-message)
  (define-key navi2ch-article-mode-map "<" 'navi2ch-article-goto-first-message)
  (define-key navi2ch-article-mode-map "\eu" 'navi2ch-article-uudecode-message)
  (define-key navi2ch-article-mode-map "f" 'navi2ch-article-forward-buffer)
  (define-key navi2ch-article-mode-map "b" 'navi2ch-article-backward-buffer)
  (define-key navi2ch-article-mode-map "1" 'navi2ch-one-pain)
  (define-key navi2ch-article-mode-map "2" 'navi2ch-article-two-pain)
  (define-key navi2ch-article-mode-map "3" 'navi2ch-three-pain)
  (define-key navi2ch-article-mode-map "\C-c\C-a" 'navi2ch-article-add-global-bookmark)
  (define-key navi2ch-article-mode-map "\C-c\C-b" 'navi2ch-bookmark))

(defvar navi2ch-article-buffer-name-prefix "*navi2ch article ")
(defvar navi2ch-article-current-article nil)
(defvar navi2ch-article-current-board nil)
(defvar navi2ch-article-message-list nil)
(defvar navi2ch-article-point-stack nil "$B0LCV$r3P$($H$/(B stack")

(defun navi2ch-article-inside-range-p (num range len)
  "NUM $B$,(B RANGE $B$G<($9HO0O$KF~$C$F$k$+(B
LEN $B$O(B RANGE $B$GHO0O$r;XDj$5$l$k(B list $B$ND9$5(B"
  (or (not range)
      (<= num (car range))
      (> num (- len (cdr range)))))

(defun navi2ch-article-parse-message (list range)
  "LIST $B$r(B RANGE $B$NHO0O$G(B parse $B$9$k!#(B
$B$9$G$K(B parse $B$5$l$F$kHO0O$O(B parse $B$7$J$$(B"
  (let ((first (car range))
        (last (cdr range))
        message-list)
    (message "parsing current messages...")
    (setq message-list
          (mapcar
           (lambda (x)
             (if (and (stringp (cdr x))
                      (navi2ch-article-inside-range-p
                       (car x) range (length list)))
                 (let ((strs (split-string (cdr x) "\\(,\\|<>\\) *"))
                       list2)
				   (when (> (length strs) 5)
					 (setq strs (split-string (cdr x) "<> *")))
				   (setq test strs)
                   (setq list2
                         (list
                          (cons 'name (nth 0 strs))
                          (cons 'mail (nth 1 strs))
                          (cons 'date (nth 2 strs))
                          (cons 'data (nth 3 strs))
                          (cons 'subject (nth 4 strs))))
                   (let ((case-fold-search nil)
                         str)
                     (dolist (elt navi2ch-article-parse-field-list)
                       (setq str (cdr (assq elt list2)))
                       (setq str
                             (funcall
                              (if (eq elt 'data)
                                  'navi2ch-replace-html-tag-with-temp-buffer
                                'navi2ch-replace-html-tag)
                              str))
                       (setcdr (assq elt list2) str)))
                   (cons (car x) list2))
               x))
           list))
    (message "parsing current messages...done")
    message-list))

(defun navi2ch-article-get-message-list (file &optional begin end)
  "FILE $B$N(B BEGIN $B$+$i(B END $B$^$G$NHO0O$+$i%9%l$N(B list $B$r:n$k(B"
  (when (file-exists-p file)
    (let ((content (with-temp-buffer
                     (navi2ch-insert-file-contents file begin end)
                     (buffer-string)))
          (i 1)
          message-list)
      (message "splitting current messages...")
      (setq message-list
            (mapcar
             (lambda (x)
               (let (elt)
                 (setq elt (cons i x))
                 (setq i (1+ i))
                 elt))
             (delete "" (split-string content "\n"))))
      (message "splitting current messages...done")
      message-list)))
                      
(defun navi2ch-article-append-message-list (list1 list2)
  (let ((num (length list1))
        list)
    (message "appending current messages...")
    (setq list (append list1
                       (mapcar
                        (lambda (x)
                          (setq num (1+ num))
                          (cons num (cdr x)))
                        list2)))
    (message "appending current messages...done")
    list))

(defun navi2ch-article-insert-messages (list range)
  "LIST $B$r@07A$7$FA^F~$9$k(B"
  (message "inserting current messages...")
  (let ((len (length navi2ch-article-message-list)))
    (dolist (x list)
      (when (navi2ch-article-inside-range-p (car x) range len)
        (setcdr x (navi2ch-put-alist 'point (point-marker) (cdr x)))
        (let ((num (car x))
              (elt (cdr x))
              (begin (point-marker)))
		  (when (cdr (assq 'data elt))
			(let ((p (point)))
			  (insert (funcall navi2ch-article-header-format-function
							   num
							   (cdr (assq 'name elt))
							   (cdr (assq 'mail elt))
							   (cdr (assq 'date elt))))
			  (put-text-property p (1+ p) 'current-number num))
			(let ((p (point)))
			  (insert (cdr (assq 'data elt)) "\n")
			  (navi2ch-article-set-link-property p (point))
			  (navi2ch-article-arrange-message p (point)))
			(let ((p (point)))
			  (insert "\n")
			  (put-text-property p (point) 'face 'underline))
			(insert "\n"))))))
  (message "inserting current messages...done"))

(defun navi2ch-article-default-header-format-function (number name mail date)
  "$B%G%U%)%k%H$N%X%C%@$r%U%)!<%^%C%H$9$k4X?t(B
$B%X%C%@$N(Bface $B$rIU$1$k$N$b$3$3$G!#(B"
  (let (str p)
    (setq str
          (concat "From: [" (number-to-string number) "] "
                  name
                  (if (string= mail "")
                      ""
                    (concat " <" mail ">"))
                  "\n")
          p (length str))
    (put-text-property 0 (1- p) 'face 'underline str)
    (setq str (concat str "Date: " date "\n\n"))
    (put-text-property p (- (length str) 2) 'face 'underline str)
    str))

(defun navi2ch-article-arrange-message (beg end)
  (let ((id (cdr (assq 'id navi2ch-article-current-board))))
    (when (or (member id navi2ch-article-enable-fill-list)
              (and (not (member id navi2ch-article-disable-fill-list))
                   navi2ch-article-enable-fill))
      (set-hard-newline-properties beg end)
      (let ((fill-column (- (window-width) 5))
            (use-hard-newlines t))
        (fill-region beg end))))
  (run-hook-with-args 'navi2ch-article-arrange-message-functions beg end))

(defun navi2ch-article-set-link-property (beg end)
  ">>1 $B$H$+(B http:// $B$K(B property $B$rIU$1$k(B"
  (save-excursion
    (save-restriction
      (narrow-to-region beg end)
      (goto-char (point-min))
      (while (re-search-forward
              "[>$B!d(B][>$B!d(B]?\\([0-9$B#0#1#2#3#4#5#6#7#8#9(B]+\\)" nil t)
        (add-text-properties
         (match-beginning 0)
         (match-end 0)
         (list 'face navi2ch-article-link-face
               'link t
               'mouse-face navi2ch-article-link-mouse-face
               'number (string-to-number
                        (japanese-hankaku (match-string 1))))))
      (goto-char (point-min))
      (while (re-search-forward
              "h?ttp://\\([-a-zA-Z0-9_=?#$@~`%&*+|\\/.,:]+\\)" nil t)
        (add-text-properties
         (match-beginning 0)
         (match-end 0)
         (list 'face navi2ch-article-link-face
               'link t
               'mouse-face navi2ch-article-link-mouse-face
               'url (concat "http://" (match-string 1))))))))
  
(defun navi2ch-article-view-article (board article &optional force)
  "$B%9%l$r8+$k!#(Bforce $B$G6/@)FI$_9~$_(B"
  (let ((buf-name (navi2ch-article-get-buffer-name board article)))
    (if (get-buffer buf-name)
        (progn
          (switch-to-buffer buf-name)
          (navi2ch-article-sync force))
      (switch-to-buffer (get-buffer-create buf-name))
      (navi2ch-article-mode)
      (setq navi2ch-article-current-board board
            navi2ch-article-current-article article)
      (when navi2ch-article-auto-range
        (if (file-exists-p (navi2ch-article-get-file-name board article))
            (setq navi2ch-article-view-range
                  navi2ch-article-exist-message-range)
          (setq navi2ch-article-view-range
                navi2ch-article-new-message-range)))
      (navi2ch-article-sync force 'first))))

(defun navi2ch-article-mode ()
"\\{navi2ch-article-mode-map}"
  (interactive)
  (setq major-mode 'navi2ch-article-mode)
  (setq mode-name "Navi2ch Article")
  (setq buffer-read-only t)
  (make-local-variable 'navi2ch-article-current-article)
  (make-local-variable 'navi2ch-article-current-board)
  (make-local-variable 'navi2ch-article-message-list)
  (make-local-variable 'navi2ch-article-point-stack)
  (make-local-variable 'truncate-partial-width-windows)
  (make-local-variable 'navi2ch-article-view-range)
  (setq truncate-partial-width-windows nil)
  (use-local-map navi2ch-article-mode-map)
  (setq navi2ch-article-point-stack nil)
  (run-hooks 'navi2ch-article-mode-hook))

(defun navi2ch-article-exit ()
  (interactive)
  ;; (navi2ch-article-add-number)
  (delete-windows-on (current-buffer))
;;  (bury-buffer navi2ch-article-buffer-name)
  (when (get-buffer navi2ch-board-buffer-name)
    (let ((win (get-buffer-window navi2ch-board-buffer-name)))
      (if win
          (select-window win)
        (switch-to-buffer navi2ch-board-buffer-name)))))

(defun navi2ch-article-kill-buffer ()
  "save-info $B$7$F$+$i(B kill-buffer $B$9$k(B"
  (interactive)
  (navi2ch-article-save-info)
  (kill-buffer (current-buffer)))

(defun navi2ch-article-sync (&optional force first)
  "$B%9%l$r99?7$9$k!#(Bforce $B$J$i6/@)!#(B
first $B$,(B nil $B$J$i$P!"%U%!%$%k$,99?7$5$l$F$J$1$l$P2?$b$7$J$$(B"
  (interactive "P")
  (let* ((list navi2ch-article-message-list)
         (range navi2ch-article-view-range)
         (article navi2ch-article-current-article)
         (board navi2ch-article-current-board)
         (navi2ch-net-force-update (or navi2ch-net-force-update
                                       force))
         (url (navi2ch-article-get-url board article))
         (file (navi2ch-article-get-file-name board article))
         (old-size (nth 7 (file-attributes file)))
         time)
    
    (when first
      (setq article (navi2ch-article-load-info)))
    (unless navi2ch-offline
      (setq time (funcall
                  (if (and (file-exists-p file)
                           navi2ch-article-enable-diff)
                      'navi2ch-net-update-file-diff
                    'navi2ch-net-update-file)
                  url file (cdr (assq 'time article))))
      (when time
        (setq article (navi2ch-put-alist 'time time article)
              navi2ch-article-current-article article)))
    (when (or first time)
      (setq list
            (if first
                (navi2ch-article-get-message-list file)
              (navi2ch-article-append-message-list
               list (navi2ch-article-get-message-list
                     file old-size))))
      (unless first
        (navi2ch-article-save-number))
      (let ((buffer-read-only nil))
        (erase-buffer)
        (setq list (navi2ch-article-parse-message list range)
              navi2ch-article-message-list list)
        (navi2ch-article-insert-messages list range))
      (navi2ch-article-load-number)
      (navi2ch-article-save-info)
      (navi2ch-article-set-mode-line)
      list)))

(defun navi2ch-article-set-mode-line ()
  (let ((article navi2ch-article-current-article)
        (list navi2ch-article-message-list))
    (unless (assq 'subject article)
      (setq list (navi2ch-article-parse-message list '(1 . 1)))
      (setq article (navi2ch-put-alist 'subject
                                       (cdr (assq 'subject (car list)))
                                       article)
            navi2ch-article-current-article article))
    (setq mode-line-buffer-identification
          (concat (cdr (assq 'subject article))
                  " (" (cdr (assq 'response article))
                  "/" (number-to-string (length list)) ")")))
  (force-mode-line-update t))

(defun navi2ch-article-sync-disable-diff (&optional force)
  (interactive "P")
  (let ((navi2ch-article-enable-diff nil))
    (navi2ch-article-sync force)))

(defun navi2ch-article-redraw ()
  "$B8=:_I=<($7$F$k%9%l$rI=<($7$J$*$9(B"
  (let ((buffer-read-only nil)
        (p (point)))
    (navi2ch-article-save-number)
    (erase-buffer)
    (setq navi2ch-article-message-list
          (navi2ch-article-parse-message
           navi2ch-article-message-list
           navi2ch-article-view-range))
    (navi2ch-article-insert-messages navi2ch-article-message-list
                                     navi2ch-article-view-range)
    (navi2ch-article-load-number)
    (goto-char p)))

(defun navi2ch-article-select-view-range-subr ()
  "$BI=<($9$kHO0O$r%-!<%\!<%I%a%K%e!<$GA*Br$9$k(B"
  (save-window-excursion
    (delete-other-windows)
    (let ((buf (get-buffer-create "*select view range*"))
          (range navi2ch-article-view-range))
      (save-excursion
        (set-buffer buf)
        (erase-buffer)
        (insert (format "   %8s %8s\n" "first" "last"))
        (insert (format "0: %17s\n" "all range"))
        (let ((i 1))
          (dolist (x navi2ch-article-view-range-list)
            (insert (format "%d: %8d %8d\n" i (car x) (cdr x)))
            (setq i (1+ i)))))
      (display-buffer buf)
      (let (n)
        (message "input: ")
        (setq n (read-char))
        (when (or (< n ?0) (> n ?9))
          (error "%c is bad key" n))
        (setq n (- n ?0))
        (setq range
              (if (eq n 0) nil
                (nth (1- n) navi2ch-article-view-range-list))))
      (kill-buffer buf)
      range)))

(defun navi2ch-article-redraw-range ()
  "$BI=<($9$kHO0O$r;XDj$7$?8e(B redraw"
  (interactive)
  (setq navi2ch-article-view-range
        (navi2ch-article-select-view-range-subr))
  (navi2ch-article-redraw))

(defun navi2ch-article-save-number ()
  (setq navi2ch-article-current-article
        (navi2ch-put-alist 'number
                           (navi2ch-article-get-current-number)
                           navi2ch-article-current-article)))

(defun navi2ch-article-load-number ()
  (let ((num (cdr (assq 'number navi2ch-article-current-article))))
    (navi2ch-article-goto-number (or num 1))))

(defun navi2ch-article-save-info ()
  (navi2ch-article-save-number)
  (let* ((article navi2ch-article-current-article)
         (board navi2ch-article-current-board)
         (alist (list
                 (assq 'number article)
                 (assq 'name article)
                 (assq 'time article))))
    (navi2ch-save-info
     (navi2ch-article-get-info-file-name board article)
     alist)))

(defun navi2ch-article-load-info ()
  (let* ((article navi2ch-article-current-article)
         (board navi2ch-article-current-board)
         (alist (navi2ch-load-info
                 (navi2ch-article-get-info-file-name board article))))
    (dolist (x alist)
      (setq article (navi2ch-put-alist (car x) (cdr x) article)))
    (setq navi2ch-article-current-article article)))

(defun navi2ch-article-write-message ()
  (interactive)
  (navi2ch-message-write-message navi2ch-article-current-board
                                 navi2ch-article-current-article))

(defun navi2ch-article-select-current-link ()
  (interactive)
  (let ((prop (get-text-property (point) 'number)))
    (if prop
        (navi2ch-article-goto-number prop t)
      (setq prop (get-text-property (point) 'url))
      (if (navi2ch-2ch-url-p prop)
          (navi2ch-article-view-article
           (navi2ch-url-to-board prop)
           (navi2ch-url-to-article prop))
        (browse-url prop)))))

(defun navi2ch-article-mouse-select (e)
  (interactive "e")
  (mouse-set-point e)
  (navi2ch-article-select-current-link))

(defun navi2ch-article-recenter (num)
  "NUM $BHVL\$N%l%9$r2hLL$N0lHV>e$K(B"
  (save-excursion
    (goto-char (cdr (assq 'point (navi2ch-article-get-message num))))
    (recenter 0)))

(defun navi2ch-article-goto-number (num &optional save)
  "NUM $BHVL\$N%l%9$K0\F0(B"
  (interactive "ninput number: ")
  (when (or (interactive-p) save)
    (navi2ch-article-push-point))
  (let ((len (length navi2ch-article-message-list))
        (range navi2ch-article-view-range))
    (unless (navi2ch-article-inside-range-p num range len)
      (setq num (1+ (- len (cdr range)))))
    (goto-char (cdr (assq 'point (navi2ch-article-get-message num)))))
  (if navi2ch-article-goto-number-recenter
      (navi2ch-article-recenter num))
  (force-mode-line-update t))

(defun navi2ch-article-pop-point ()
  "stack $B$+$i(B pop $B$7$?0LCV$K0\F0$9$k(B"
  (interactive)
  (let ((point (pop navi2ch-article-point-stack)))
    (if point
        (goto-char point)
      (message "stack is empty"))))

(defun navi2ch-article-push-point (&optional point)
  "$B8=:_0LCV$+(B POINT $B$r(B stack $B$K(B push $B$9$k(B"
  (interactive)
  (push (or point (point)) navi2ch-article-point-stack)
  (message "push current point"))

(defun navi2ch-article-goto-last-message ()
  "$B:G8e$N%l%9$X(B"
  (interactive)
  (navi2ch-article-goto-number
   (save-excursion
     (goto-char (point-max))
     (navi2ch-article-get-current-number))) t)

(defun navi2ch-article-goto-first-message ()
  "$B:G=i$N%l%9$X(B"
  (interactive)
  (navi2ch-article-goto-number
   (save-excursion
     (goto-char (point-min))
     (navi2ch-article-get-current-number))) t)

(defun navi2ch-article-scroll-up ()
  (interactive)
  (condition-case error
      (scroll-up)
    (end-of-buffer
     (navi2ch-article-through-next)))
  (force-mode-line-update t))

(defun navi2ch-article-scroll-down ()
  (interactive)
  (condition-case error
      (scroll-down)
    (beginning-of-buffer
     (navi2ch-article-through-prev)))
  (force-mode-line-update t))

(defun navi2ch-article-through-next ()
  (interactive)
  (when (or navi2ch-article-enable-through
            (interactive-p))
    (if (get-buffer-window navi2ch-board-buffer-name)
        (select-window (get-buffer-window navi2ch-board-buffer-name))
      (switch-to-buffer navi2ch-board-buffer-name))
    (forward-line)
    (navi2ch-board-select-article)))

(defun navi2ch-article-through-prev ()
  (interactive)
  (when (or navi2ch-article-enable-through
            (interactive-p))
    (if (get-buffer-window navi2ch-board-buffer-name)
        (select-window (get-buffer-window navi2ch-board-buffer-name))
      (switch-to-buffer navi2ch-board-buffer-name))
    (forward-line -1)
    (navi2ch-board-select-article)))

(defun navi2ch-article-two-pain ()
  (interactive)
  (let* ((list-buf (get-buffer navi2ch-list-buffer-name))
         (board-buf (get-buffer navi2ch-board-buffer-name))
         (art-buf (navi2ch-article-current-buffer))
         (list-win (get-buffer-window (or list-buf "")))
         (board-win (get-buffer-window (or board-buf "")))
         buf)
    (when art-buf
      (delete-other-windows)
      (switch-to-buffer art-buf)
      (setq buf
            (cond ((and list-buf board-buf)
                   (cond ((and list-win board-win) board-buf)
                         (list-win board-buf)
                         (board-win list-buf)
                         (t board-buf)))
                  (list-buf list-buf)
                  (board-buf board-buf)))
      (when buf
        (if (eq buf list-buf)
            (split-window-horizontally navi2ch-list-window-width)
          (split-window-vertically navi2ch-board-window-height))
        (switch-to-buffer buf))
      (other-window 1))))

(defun navi2ch-article-get-message (num)
  "NUM $BHVL\$N%l%9$rF@$k(B"
  (cdr (assq num navi2ch-article-message-list)))

(defun navi2ch-article-get-current-number ()
  "$B:#$N0LCV$N%l%9$NHV9f$rF@$k(B"
  (or (get-text-property (point) 'current-number)
      (let ((pos (previous-single-property-change
                  (point)
                  'current-number)))
	(and pos
	     (or (get-text-property pos 'current-number)
			 (get-text-property (1- pos) 'current-number))))))

(defun navi2ch-article-show-url ()
  "url $B$rI=<($7$F!"$=$N(B url $B$r8+$k$+(B kill ring $B$K%3%T!<$9$k(B"
  (interactive)
  (let ((url (navi2ch-article-to-url navi2ch-article-current-board
									 navi2ch-article-current-article)))
	(message "c)opy v)iew? URL: %s" url)
	(let ((char (read-char)))
	  (funcall (cond ((eq char ?c) 'kill-new)
					 ((eq char ?v) 'browse-url))
			   (navi2ch-article-show-url-subr)))))

(defun navi2ch-article-show-url-subr ()
  "$B%a%K%e!<$rI=<($7$F!"(Burl $B$rF@$k(B"
  (message "a)ll c)rrent r)egion b)oard")
  (let ((char (read-char)))
	(if (eq char ?b)
		(navi2ch-board-to-url navi2ch-article-current-board)
	  (apply 'navi2ch-article-to-url
			 navi2ch-article-current-board navi2ch-article-current-article
			 (cond ((eq char ?a) nil)
				   ((eq char ?c) (list (navi2ch-article-get-current-number)
									   (navi2ch-article-get-current-number)
									   t))
				   ((eq char ?r) (list (save-excursion
										 (goto-char (region-beginning))
										 (navi2ch-article-get-current-number))
									   (save-excursion
										 (goto-char (region-end))
										 (navi2ch-article-get-current-number)))))))))

(defun navi2ch-article-redisplay-current-message ()
  "$B:#$$$k%l%9$r2hLL$NCf?4(B($B>e!)(B)$B$K(B"
  (interactive)
  (navi2ch-article-recenter
   (navi2ch-article-get-current-number)))

(defun navi2ch-article-next-message ()
  "$B<!$N%a%C%;!<%8$X(B"
  (interactive)
  (let ((num (1+ (navi2ch-article-get-current-number)))
        (len (length navi2ch-article-message-list))
        (range navi2ch-article-view-range))
    (condition-case error
        (navi2ch-article-goto-number
         (if (navi2ch-article-inside-range-p num range len)
             num
           (1+ (- len (cdr range)))))
      (error
       (navi2ch-article-through-next)))))

(defun navi2ch-article-previous-message ()
  "$BA0$N%a%C%;!<%8$X(B"
  (interactive)
  (let ((num (1- (navi2ch-article-get-current-number)))
        (len (length navi2ch-article-message-list))
        (range navi2ch-article-view-range))
    (condition-case error
        (navi2ch-article-goto-number
         (if (navi2ch-article-inside-range-p num range len)
             num
           (car range)))
      (error
       (navi2ch-article-through-prev)))))

(defun navi2ch-article-next-link ()
  "$B<!$N%j%s%/$X(B"
  (interactive)
  (let ((point (or (and (get-text-property (point) 'link)
                        (next-single-property-change (point) 'link))
                   (point))))
    (setq point (next-single-property-change point 'link))
    (and point (goto-char point))))

(defun navi2ch-article-previous-link ()
  "$BA0$N%j%s%/$X(B"
  (interactive)
  (let ((point (or (and (get-text-property (point) 'link)
                        (previous-single-property-change (point) 'link))
                   (point))))
    (setq point (previous-single-property-change point 'link))
    (or (get-text-property point 'link)
        (setq point (previous-single-property-change point 'link)))
    (and point (goto-char point))))

(defun navi2ch-article-uudecode-message ()
  (interactive)
  (with-temp-buffer
    (insert (cdr
             (assq 'data
                   (save-excursion
                     (set-buffer (navi2ch-article-current-buffer))
                     (navi2ch-article-get-message
                      (navi2ch-article-get-current-number))))))
    (navi2ch-uudecode-region (point-min) (point-max))))
  
(defun navi2ch-article-add-global-bookmark ()
  (interactive)
  (navi2ch-bookmark-add
   navi2ch-article-current-board
   navi2ch-article-current-article))

(defun navi2ch-article-get-buffer-name (board article)
  (concat navi2ch-article-buffer-name-prefix
          (cdr (assq 'id board))
          "/"
          (cdr (assq 'artid article))))
                    
(defun navi2ch-article-buffer-list ()
  "`navi2ch-article-mode' $B$N(B buffer $B$N(B list $B$rJV$9(B"
  (let (list)
    (dolist (x (buffer-list))
      (when (save-excursion
              (set-buffer x)
              (eq major-mode 'navi2ch-article-mode))
        (setq list (cons x list))))
    (nreverse list)))

(defun navi2ch-article-current-buffer ()
  "`buffer-list' $B$N0lHV:G=i$N(B `navi2ch-article-mode' $B$N(B buffer $B$rJV$9(B"
  (let ((list (buffer-list)))
    (catch 'loop
      (while list
        (when (save-excursion
                (set-buffer (car list))
                (eq major-mode 'navi2ch-article-mode))
          (throw 'loop (car list)))
        (setq list (cdr list)))
      nil)))

(defun navi2ch-article-forward-buffer ()
  "$B<!$N(B article buffer $B$X(B"
  (interactive)
  (let (buf)
    (dolist (x (buffer-list))
      (when (save-excursion
              (set-buffer x)
              (eq major-mode 'navi2ch-article-mode))
        (setq buf x)))
    (switch-to-buffer buf)))
  
(defun navi2ch-article-backward-buffer ()
  "$BA0$N(B article buffer $B$X(B"
  (interactive)
  (bury-buffer)
  (switch-to-buffer (navi2ch-article-current-buffer)))


(defun navi2ch-article-get-url (board article)
  (concat (cdr (assq 'uri board))
          "dat/" (cdr (assq 'artid article))
          ".dat"))

(defun navi2ch-article-get-file-name (board article)
  (navi2ch-expand-file-name
   (let ((uri (cdr (assq 'uri board))))
     (string-match "http://\\(.+\\)" uri)
     (concat (match-string 1 uri) (cdr (assq 'artid article)) ".dat"))))

(defun navi2ch-article-get-info-file-name (board article)
  (navi2ch-expand-file-name
   (let ((uri (cdr (assq 'uri board))))
     (string-match "http://\\(.+\\)" uri)
     (concat (match-string 1 uri) "info/" (cdr (assq 'artid article))))))

(defsubst navi2ch-article-check-cached (board article)
  "BOARD $B$H(B ARTICLE $B$G;XDj$5$l$k%9%l%C%I$,%-%c%C%7%e$5$l$F$k$+!#(B"
  (cond ((get-buffer (navi2ch-article-get-buffer-name board article))
         'view)
        ((file-exists-p (navi2ch-article-get-file-name board article))
          'cache)))

(provide 'navi2ch-article)

;;; navi2ch-article.el ends here