;;; navi2ch-net.el --- Network module for navi2ch

;; Copyright (C) 2000 by 2$B$A$c$s$M$k(B

;; Author: (not 1)
;; Keywords: network 2ch

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(require 'timezone)

(defvar navi2ch-net-connection-name "navi2ch connection")
(defvar navi2ch-net-coding-system 'shift_jis)
(defvar navi2ch-net-http-proxy
  (if (string= (getenv "HTTP_PROXY") "")
      nil
    (getenv "HTTP_PROXY")))
(defvar navi2ch-net-force-update nil)
(defvar navi2ch-net-check-margin 100)
(defvar navi2ch-net-turn-back-step 1000)
(defvar navi2ch-net-turn-back-when-aborn t)

(defun navi2ch-net-send-request (url method &optional other-header content)
  (let ((buf (get-buffer-create (concat " *" navi2ch-net-connection-name)))
        (process-connection-type nil)
        host file port)
    (let ((list (navi2ch-net-split-url url navi2ch-net-http-proxy)))
      (setq host (cdr (assq 'host list))
            file (cdr (assq 'file list))
            port (cdr (assq 'port list))
	    host2ch (cdr (assq 'host2ch list))))
    (save-excursion
        (set-buffer buf)
        (erase-buffer))
    (message "now connecting...")
    (let ((proc (open-network-stream
                 navi2ch-net-connection-name buf host port)))
      (set-process-coding-system proc 'binary 'binary)
      (set-process-sentinel proc (lambda (p s))) ; exited abnormary $B$r=P$5$J$/$9$k(B
      (process-send-string
       proc
       (format (concat
                "%s %s HTTP/1.1\r\n"
                "MIME-Version: 1.0\r\n"
                "Host: %s\r\n"
                "Connection: close\r\n"
                "%s"                    ;other-header
                "%s"                    ;content
                "\r\n")
               method file
               host2ch
               (if other-header
                   (concat (mapconcat
                            (function
                             (lambda (x)
                               (concat (car x) ": " (cdr x))))
                            (delq nil other-header) "\r\n")
                           "\r\n")
                 "")
               (if content
                   (format "Content-length: %d\r\n\r\n%s"
                           (length content) content)
                 "")))
      (message "now connecting...connected")
      proc)))
      
(defun navi2ch-net-split-url (url &optional proxy)
  (let (host file port host2ch)
    (string-match "http://\\([^/]+\\)" url)    
    (setq host2ch (match-string 1 url))
    (if proxy
        (progn
          (string-match "^\\(http://\\)?\\(.*\\):\\([0-9]+\\)" proxy)
          (list
           (cons 'host (match-string 2 proxy))
           (cons 'file url)
           (cons 'port (string-to-number (match-string 3 proxy)))
           (cons 'host2ch host2ch)))
      (string-match "http://\\([^/]+\\)\\(.*\\)" url)
      (list
       (cons 'host (match-string 1 url))
       (cons 'file (match-string 2 url))
       (cons 'port  80)
       (cons 'host2ch host2ch)))))

(defun navi2ch-net-get-status (proc)
  "`proc' $B$N@\B3$N%9%F!<%?%9It$rJV$9(B"
  (save-excursion
    (set-buffer (process-buffer proc))
    (while (and (eq (process-status proc) 'open)
                (goto-char (point-min))
                (not (looking-at "HTTP/1\\.[01] \\([0-9]+\\)")))
      (accept-process-output proc))
    (match-string 1)))

(defun navi2ch-net-get-header (proc)
  "`proc' $B$N@\B3$N%X%C%@It$rJV$9(B"
  (save-excursion
    (set-buffer (process-buffer proc))
    (while (and (eq (process-status proc) 'open)
                (goto-char (point-min))
                (not (re-search-forward "\r\n\r?\n" nil t)))
      (accept-process-output proc))
    (goto-char (point-min))
    (save-excursion
      (re-search-forward "\r\n\r?\n"))
    (let ((end (match-end 0))
          list)
      (while (re-search-forward "^\\([^\r\n:]+\\): \\(.+\\)\r\n" end t)
        (setq list (cons (cons (match-string 1) (match-string 2)) 
                         list)))
      (nreverse list))))

(defun navi2ch-net-get-content (proc)
  "`proc' $B$N@\B3$NK\J8$rJV$9(B"
  (save-excursion
    (set-buffer (process-buffer proc))
    (while (eq (process-status proc) 'open)
      (accept-process-output proc))
    (goto-char (point-min))
    (re-search-forward "\r\n\r?\n")
    (buffer-substring (point) (point-max))))

(defun navi2ch-net-download-file (url &optional time accept-status other-header)
  "`url' $B$+$i%@%&%s%m!<%I$r3+;O$9$k!#(B
`time' $B$,(B `non-nil' $B$J$i$P(B `time' $B$h$j?7$7$$;~$@$1%@%&%s%m!<%I$9$k!#(B
$B%j%9%H(B `accept-status' $B$,(B `non-nil' $B$J$i$P%9%F!<%?%9$,(B `accept-status' $B$K4^$^$l(B
$B$F$$$k;~$@$1%@%&%s%m!<%I$9$k!#(B
`other-header' $B$,(B `non-nil' $B$J$i$P%j%/%(%9%H$K$3$N%X%C%@$rDI2C$9$k!#(B
$B%@%&%s%m!<%I$G$-$l$P$=$N@\B3$rJV$9!#(B"
  (let ((proc
         (navi2ch-net-send-request 
          url "GET"
          (append
           (list (if navi2ch-net-force-update
                     (cons "Pragma" "no-cache")
                   (when time
                     (cons "If-Modified-Since" time))))
           other-header))))
    (message "checking file is updated...")
    (let ((status (navi2ch-net-get-status proc)))
      (cond ((string= status "404")
             (message "file is not found")
             (delete-process proc)
             nil)
            ((or (string= status "304")
                 (and accept-status
                      (not (member status accept-status))))
             (message "file is not updated")
             (delete-process proc)
             nil)
            (t
             (message "file is updated")
             proc)))))

(defun navi2ch-net-download-file-range (url range &optional time)
  "Range $B%X%C%@$r;H$C$F%U%!%$%k$r%@%&%s%m!<%I$9$k!#(B"
  (navi2ch-net-download-file url time '("206")
                             (list (cons "Range" (concat "bytes=" range)))))
  
  
(defun navi2ch-net-update-file (url file &optional time func)
  "`file' $B$r99?7$9$k!#(B
`time' $B$,(B `non-nil' $B$J$i$P(B `time' $B$h$j?7$7$$;~$@$199?7$9$k!#(B
$B99?7$G$-$l$P99?7F|;~$rJV$9(B"
  (let ((dir (file-name-directory file)))
    (unless (file-exists-p dir)
      (make-directory dir t)))
  (let ((proc (navi2ch-net-download-file url time '("200"))))
    (if proc
        (progn
          (message "%s getting new file..." (current-message))
          (let* ((coding-system-for-write 'binary)
                 (coding-system-for-read 'binary)
                 (cont (navi2ch-net-get-content proc)))
            (with-temp-file file 
              (insert (funcall (if func (progn (message "translating...")
                                               func)
                                 'eval)
                               cont)))
            (message "%sdone" (current-message))
            ;; $B$I$&$9$l$P@dBP$K(B Last-Modified $B$rLc$($k$+$J!)(B
            (or (cdr (assoc "Last-Modified"
                            (navi2ch-net-get-header proc)))
                t)))
      nil)))

(defun navi2ch-net-update-file-diff (url file &optional time)
  "file $B$r:9J,$G99?7$9$k!#(B
`time' $B$,(B `non-nil' $B$J$i$P(B `time' $B$h$j?7$7$$;~$@$199?7$9$k!#(B
$B99?7$G$-$l$P99?7F|;~$rJV$9(B"
  (let ((dir (file-name-directory file)))
    (unless (file-exists-p dir)
      (make-directory dir t)))
  (let* ((size (- (nth 7 (file-attributes file))
                  navi2ch-net-check-margin))
         (proc (navi2ch-net-download-file-range
                url (format "%d-" size) time)))
    (if proc
        (progn
          (message "%s getting file diff..." (current-message))
          (let* ((coding-system-for-write 'binary)
                 (coding-system-for-read 'binary)
                 (cont (navi2ch-net-get-content proc)))
            (if (navi2ch-net-check-file file cont)
                (progn
                  (with-temp-file file 
                    (insert-file-contents file nil nil size)
                    (goto-char (point-max))
                    (insert cont))
                  (message "%sdone" (current-message))
                  (cdr (assoc "Last-Modified"
                              (navi2ch-net-get-header proc))))
              (message "$B$"$\!<$s(B!!!")
              (or (and navi2ch-net-turn-back-when-aborn
                       (navi2ch-net-turn-back-file url file))
                  (navi2ch-net-update-file url file)))))
      nil)))

(defun navi2ch-net-check-file (file cont)
  (let (str)
    (with-temp-buffer
      (insert-file-contents file nil
                            (- (nth 7 (file-attributes file))
                               navi2ch-net-check-margin))
      (setq str (buffer-string)))
    (string= (string-as-unibyte str)
             (substring (string-as-unibyte cont) 0 navi2ch-net-check-margin))))

(defun navi2ch-net-turn-back-file (url file)
  (let* ((size (- (nth 7 (file-attributes file))
                  navi2ch-net-turn-back-step))
         (proc (navi2ch-net-download-file-range
                url (format "%d-" size) time)))
    (message "$B$"$\!<$s(B!!! turn back and searching same string...")
    (let ((cont (navi2ch-net-get-content proc)))
      (with-temp-buffer
        (set-buffer-multibyte nil)
        (insert-file-contents file)
        (goto-char (point-max))
        (if (search-backward (string-as-unibyte cont) nil t)
            (progn
              (goto-char (match-beginning 0))
              (delete-region (point) (point-max))
              (insert cont)
              (message "%sdone" (current-message))
              (cdr (assoc "Last-Modified"
                          (navi2ch-net-get-header proc))))
          (message "not found same string")
          nil)))))
  
;; from Emacs/W3
(defconst navi2ch-net-url-unreserved-chars
  '(
    ?a ?b ?c ?d ?e ?f ?g ?h ?i ?j ?k ?l ?m ?n ?o ?p ?q ?r ?s ?t ?u ?v ?w ?x ?y ?z
    ?A ?B ?C ?D ?E ?F ?G ?H ?I ?J ?K ?L ?M ?N ?O ?P ?Q ?R ?S ?T ?U ?V ?W ?X ?Y ?Z
    ?0 ?1 ?2 ?3 ?4 ?5 ?6 ?7 ?8 ?9
    ?$ ?- ?_ ?. ?! ?~ ?* ?' ?\( ?\) ?,)
  "A list of characters that are _NOT_ reserve in the URL spec.
This is taken from draft-fielding-url-syntax-02.txt - check your local
internet drafts directory for a copy.")

;; from Emacs/W3
(defun navi2ch-net-url-hexify-string (str)
  "Escape characters in a string"
  (mapconcat
   (function
    (lambda (char)
      (if (not (memq char navi2ch-net-url-unreserved-chars))
   (if (< char 16)
       (upcase (format "%%0%x" char))
     (upcase (format "%%%x" char)))
   (char-to-string char))))
   (encode-coding-string str navi2ch-net-coding-system) ""))

(defun navi2ch-net-get-param-string (param-alist)
  (mapconcat
   (function
    (lambda (x)
      (concat (car x) "=" (navi2ch-net-url-hexify-string (cdr x)))))
   param-alist
   "&"))

(defun navi2ch-net-send-message (from mail message subject board article)
  (let ((param-alist
         (list
          (cons "SUBMIT" "$B=q$-9~$`(B")
          (cons "FROM" (or from ""))
          (cons "mail" (or mail ""))
          (cons "bbs" (cdr (assq 'id board)))
          (cons "time"
                (mapconcat 'int-to-string
                           (let ((time (current-time)))
                             (navi2ch-bigint-add
                              (navi2ch-bigint-multiply
                               (nth 0 time) (expt 2 16)) (nth 1 time)))
                           ""))
          (cons "MESSAGE" message)
          (if subject
              (cons "subject" subject)
            (cons "key" (cdr (assq 'artid article)))))))
    (let ((uri (cdr (assq 'uri board)))
          proc)
      (string-match "\\(http://.+\\)/[^/]+" uri)
      (setq uri (match-string 1 uri))
      (setq proc (navi2ch-net-send-request
                  (concat uri "/test/bbs.cgi")
                  "POST"
                  (list (cons "Content-Type"
                              "application/x-www-form-urlencoded")
                        (cons "Cookie"
                              (concat "NAME=" from
                                      "; MAIL=" mail))
                        (cons "Referer"
                              (concat uri "/"
                                      (cdr (assq 'id board)) "/"
                                      "index2.html")))
                  (navi2ch-net-get-param-string param-alist)))
      (message "send message...")
      (if (string= (navi2ch-net-get-status proc) "302")
          (progn
            (message "send message...succeed")
            (delete-process proc) t)
        (message "send message...failed")
        (delete-process proc) nil))))

(provide 'navi2ch-net)